#include "PLATFORM.H"
#include "DPMI/DBGUTIL.H"
#include "SBEMU.H"

#define SBEMU_RESET_START 0
#define SBEMU_RESET_END 1
#define SBEMU_RESET_POLL 2
void(*SBEMU_StartCB)(void);
static int SBEMU_ResetState = SBEMU_RESET_END;
static int SBEMU_Started = 0;
static int SBEMU_IRQ = 5;
static int SBEMU_DMA = 1;
static int SBEMU_DACSpeaker = 1;
static int SBEMU_Bits = 8;
static int SBEMU_SampleRate = 22050;
static int SBEMU_Samples = 0;
static int SBEMU_Auto = 0;
static int SBEMU_HighSpeed = 0;
static int SBEMU_DSPCMD = 0;
static int SBEMU_DSPCMD_Subindex = 0;
static int SBEMU_DSPDATA_Subindex = 0;
static int SBEMU_TriggerIRQ = 0;
static int SBEMU_Pos = 0;
static uint8_t SBEMU_IRQMap[4] = {2,5,7,10};
static uint8_t SBEMU_MixerRegIndex = 0;
static uint8_t SBEMU_idbyte;
static uint8_t SBEMU_WS;
static int SBEMU_TimeConstantMapMono[][2] =
{
    0xA5, 11025,
    0xD2, 22050,
    0xE9, 44100,
};
static uint8_t SBEMU_Copyright[] = "COPYRIGHT (C) CREATIVE TECHNOLOGY LTD, 1992.";

static uint8_t SBEMU_MixerRegs[256];

static int SBEMU_Indexof(uint8_t* array, int count, uint8_t  val)
{
    for(int i = 0; i < count; ++i)
    {
        if(array[i] == val)
            return i;
    }
    return -1;
}


void SBEMU_Mixer_WriteAddr(int16_t port, uint8_t value)
{
    _LOG("SBEMU: mixer wirte addr: %x %x\n", port, value);
    SBEMU_MixerRegIndex = value;
}

void SBEMU_Mixer_Write(uint16_t port, uint8_t value)
{
    _LOG("SBEMU: mixer wirte: %x\n", value);
    SBEMU_MixerRegs[SBEMU_MixerRegIndex] = value;
}

uint8_t SBEMU_Mixer_Read(uint16_t port)
{
    _LOG("SBEMU: mixer read: %x\n", SBEMU_MixerRegs[SBEMU_MixerRegIndex]);
    return SBEMU_MixerRegs[SBEMU_MixerRegIndex];
}

void SBEMU_DSP_Reset(uint16_t port, uint8_t value)
{
    _LOG("SBEMU: DSP reset: %d\n",value);
    if(value == 1)
    {
        SBEMU_ResetState = SBEMU_RESET_START;
        SBEMU_MixerRegs[SBEMU_MIXERREG_INT_SETUP] = 1<<SBEMU_Indexof(SBEMU_IRQMap,countof(SBEMU_IRQMap),SBEMU_IRQ);
        SBEMU_MixerRegs[SBEMU_MIXERREG_DMA_SETUP] = (1<<SBEMU_DMA)&0xEB;
        SBEMU_MixerRegs[SBEMU_MIXERREG_MODEFILTER] = 0xFD; //mask out stereo
        SBEMU_MixerRegIndex = 0;
        SBEMU_DSPCMD = 0;
        SBEMU_DSPCMD_Subindex = 0;
        SBEMU_DSPDATA_Subindex = 0;
        SBEMU_Started = 0;
        SBEMU_Samples = 0;
        SBEMU_Auto = 0;
        SBEMU_Bits = 8;
        SBEMU_Pos = 0;
        SBEMU_HighSpeed = 0;
        SBEMU_TriggerIRQ = 0;
    }
    if(value == 0 && SBEMU_ResetState == SBEMU_RESET_START)
        SBEMU_ResetState = SBEMU_RESET_POLL;
}

void SBEMU_DSP_Write(uint16_t port, uint8_t value)
{
    _LOG("SBEMU: DSP write %02x, original: %02x\n", value, SBEMU_DSPCMD);
    if(SBEMU_HighSpeed) //highspeed won't accept further commands, need reset
        return;
    int OldStarted = SBEMU_Started;
    if(SBEMU_DSPCMD == 0)
    {
        SBEMU_DSPCMD = value;
        SBEMU_DSPCMD_Subindex = 0;
        switch(SBEMU_DSPCMD)
        {
            case SBEMU_CMD_TRIGGER_IRQ:
            {
                SBEMU_TriggerIRQ = 1;
                SBEMU_DSPCMD = 0;
                //VIRQ_Invoke(SBEMU_GetIRQ()); //not working
            }
            break;
            case SBEMU_CMD_DAC_SPEAKER_ON:
            case SBEMU_CMD_DAC_SPEAKER_OFF:
                SBEMU_DSPCMD = 0;
                break;
            case SBEMU_CMD_HALT_DMA:
            case SBEMU_CMD_CONTINUE_DMA:
            {
                SBEMU_Started = SBEMU_DSPCMD == SBEMU_CMD_CONTINUE_DMA;
                SBEMU_DSPCMD = 0;
            }
            break;
            case SBEMU_CMD_8BIT_OUT_AUTO_HS:
            case SBEMU_CMD_8BIT_OUT_AUTO:
            {
                SBEMU_Auto = TRUE;
                SBEMU_Bits = 8;
                SBEMU_HighSpeed = (SBEMU_DSPCMD==SBEMU_CMD_8BIT_OUT_AUTO_HS);
                SBEMU_Started = TRUE; //start transfer
                SBEMU_DSPCMD = 0;
                SBEMU_Pos = 0;
            }
            break;
            case SBEMU_CMD_EXIT_16BIT_AUTO:
            case SBEMU_CMD_EXIT_8BIT_AUTO:
            {
                if(SBEMU_Auto)
                {
                    SBEMU_Auto = FALSE;
                    SBEMU_Started = FALSE;
                }
                SBEMU_DSPCMD = 0;
            }
            break;
        }
    }
    else
    {
        switch(SBEMU_DSPCMD)
        {
            case SBEMU_CMD_SET_TIMECONST:
            {
                SBEMU_SampleRate = 0;
                for(int i = 0; i < 3; ++i)
                {
                    if(value >= SBEMU_TimeConstantMapMono[i][0]-3 && value <= SBEMU_TimeConstantMapMono[i][0]+3)
                    {
                        SBEMU_SampleRate = SBEMU_TimeConstantMapMono[i][1] / SBEMU_GetChannels();
                        break;
                    }
                }
                if(SBEMU_SampleRate == 0)
                    SBEMU_SampleRate = 256000000/(65536-(value<<8)) / SBEMU_GetChannels();
                SBEMU_DSPCMD_Subindex = 2; //only 1byte
                //_LOG("SBEMU: set sampling rate: %d", SBEMU_SampleRate);
            }
            break;
            case SBEMU_CMD_SET_SIZE: //used for auto command
            case SBEMU_CMD_8BIT_OUT_1_HS:
            case SBEMU_CMD_8BIT_OUT_1:
            {
              if(SBEMU_DSPCMD_Subindex++ == 0)
                    SBEMU_Samples = value;
                else
                {
                    SBEMU_Samples |= value<<8;
                    SBEMU_Started = (SBEMU_DSPCMD==SBEMU_CMD_8BIT_OUT_1 || SBEMU_DSPCMD==SBEMU_CMD_8BIT_OUT_1_HS); //start transfer
                    SBEMU_HighSpeed = (SBEMU_DSPCMD==SBEMU_CMD_8BIT_OUT_AUTO_HS);
                    if(SBEMU_Started)
                    {
                        SBEMU_Bits = 8;
                        SBEMU_Pos = 0;
                    }
                }
            }
            break;
            case SBEMU_CMD_8OR16_OUT_1: //command start: sample rate, next command: mode
            {
                if(SBEMU_DSPCMD_Subindex++ == 0)
                    SBEMU_SampleRate = value<<8;
                else
                {
                    SBEMU_SampleRate &= ~0xFF;
                    SBEMU_SampleRate |= value;
                }
            }
            break;
            case SBEMU_CMD_8OR16_8_OUT_1:
            case SBEMU_CMD_8OR16_8_OUT_AUTO:
            case SBEMU_CMD_8OR16_8_OUT_AUTO_NOFIFO:
            case SBEMU_CMD_8OR16_16_OUT_1:
            case SBEMU_CMD_8OR16_16_OUT_AUTO:
            {
                SBEMU_Auto = (SBEMU_DSPCMD==SBEMU_CMD_8OR16_8_OUT_AUTO || SBEMU_DSPCMD==SBEMU_CMD_8OR16_16_OUT_AUTO || SBEMU_DSPCMD==SBEMU_CMD_8OR16_8_OUT_AUTO_NOFIFO);
                SBEMU_DSPCMD = value; //set next command: mode
                SBEMU_DSPCMD_Subindex = 0;
            }
            break;
            case SBEMU_CMD_MODE_PCM8_MONO:
            case SBEMU_CMD_MODE_PCM8_STEREO:
            case SBEMU_CMD_MODE_PCM16_MONO:
            case SBEMU_CMD_MODE_PCM16_STEREO:
            {
                if(SBEMU_DSPCMD_Subindex++ == 0)
                    SBEMU_Samples = value;
                else
                {
                    SBEMU_Samples |= value<<8;
                    SBEMU_Bits = (SBEMU_DSPCMD==SBEMU_CMD_MODE_PCM8_MONO || SBEMU_DSPCMD==SBEMU_CMD_MODE_PCM8_STEREO) ? 8 : 16;
                    SBEMU_MixerRegs[SBEMU_MIXERREG_MODEFILTER] |= (SBEMU_DSPCMD==SBEMU_CMD_MODE_PCM8_STEREO || SBEMU_DSPCMD==SBEMU_CMD_MODE_PCM16_STEREO) ? 2 : 0;
                    SBEMU_Started = TRUE; //start transfer here
                    SBEMU_Pos = 0;
                }
            }
            break;
            case SBEMU_CMD_DSP_ID:
            {
                SBEMU_idbyte = value;
            }
            break;
        }
        if(SBEMU_DSPCMD_Subindex >= 2)
            SBEMU_DSPCMD = 0;
    }
    if(SBEMU_Started && !OldStarted)//handle driver detection
    {
        /*if(SBEMU_StartCB)
        {
            CLIS();
            SBEMU_StartCB(); //if don't do this, need always output 0 PCM to keep interrupt alive for now
            STIL();
        }*/
    }
}

uint8_t SBEMU_DSP_Read(uint16_t port)
{
    _LOG("SBEMU: DSP read.\n");
    if(SBEMU_ResetState == SBEMU_RESET_POLL || SBEMU_ResetState == SBEMU_RESET_START)
    {
		SBEMU_ResetState = SBEMU_RESET_END;
        //_LOG("SBEMU: DSP reset read.\n");
        return 0xAA; //reset ready
    }
    if(SBEMU_DSPCMD == SBEMU_CMD_DSP_GETVER)
    {
        //https://github.com/joncampbell123/dosbox-x/wiki/Hardware:Sound-Blaster:DSP-commands:0xE1
        if(SBEMU_DSPDATA_Subindex++ == 0)
            return 0x03;    //3.2. TODO: 4.0 (SB16) support
        else
        {
            SBEMU_DSPDATA_Subindex = 0;
            SBEMU_DSPCMD = 0;
            //_LOG("SBEMU: DSP get version.\n");
            return 0x02;
        }
    }
    else if(SBEMU_DSPCMD == SBEMU_CMD_DSP_ID)
    {
        SBEMU_DSPCMD = 0;
        SBEMU_DSPDATA_Subindex = 0;
        return SBEMU_idbyte^0xFF;
    }
    else if(SBEMU_DSPCMD == SBEMU_CMD_DSP_COPYRIGHT)
    {
        if(SBEMU_DSPDATA_Subindex == sizeof(SBEMU_Copyright)-1)
        {
            SBEMU_DSPCMD = 0;
            SBEMU_DSPDATA_Subindex = 0;
            return 0xFF;
        }
        return SBEMU_Copyright[SBEMU_DSPDATA_Subindex++];
    }
    return 0xFF;
}

uint8_t SBEMU_DSP_WriteStatus(uint16_t port)
{
    _LOG("SBEMU: DSP WS\n");
    //return 0; //ready for write (bit7 clear)
    SBEMU_WS += 0x80; //some games will wait on busy first
    return SBEMU_WS;
}

uint8_t SBEMU_DSP_ReadStatus(uint16_t port)
{
    _LOG("SBEMU: DSP RS\n");
    if(SBEMU_ResetState == SBEMU_RESET_POLL || SBEMU_ResetState == SBEMU_RESET_START
    || SBEMU_DSPCMD == SBEMU_CMD_DSP_GETVER
    || SBEMU_DSPCMD == SBEMU_CMD_DSP_ID
    || SBEMU_DSPCMD == SBEMU_CMD_DSP_COPYRIGHT)
        return 0xFF; //ready for read (bit7 set)
    SBEMU_MixerRegs[SBEMU_MIXERREG_INT_STS] &= ~0x1;
    SBEMU_TriggerIRQ = 0;
    return 0x00;
}

uint8_t SBEMU_DSP_INT16ACK(uint16_t port)
{
    SBEMU_MixerRegs[SBEMU_MIXERREG_INT_STS] &= ~0x2;
    return 0xFF;
}

void SBEMU_Init(int irq, int dma, void(*startCB)(void))
{
    SBEMU_IRQ = irq;
    SBEMU_DMA = dma;
    SBEMU_StartCB = startCB;
}

uint8_t SBEMU_GetIRQ()
{
    if(SBEMU_MixerRegs[SBEMU_MIXERREG_INT_SETUP] == 0)
        return 0xFF;
    int bit = BSF(SBEMU_MixerRegs[SBEMU_MIXERREG_INT_SETUP]);
    if(bit >= 4)
        return 0xFF;
    return SBEMU_IRQMap[bit];
}

uint8_t SBEMU_GetDMA()
{
    if(SBEMU_MixerRegs[SBEMU_MIXERREG_DMA_SETUP] == 0)
        return 0xFF;
    int bit = BSF(SBEMU_MixerRegs[SBEMU_MIXERREG_DMA_SETUP]);
    return bit;
}

int SBEMU_HasStarted()
{
    return SBEMU_Started;
}

void SBEMU_Stop()
{
    SBEMU_Started = FALSE;
    SBEMU_HighSpeed = FALSE;
    SBEMU_Pos = 0;
}

int SBEMU_GetDACSpeaker()
{
    return SBEMU_DACSpeaker;
}

int SBEMU_GetBits()
{
    return SBEMU_Bits;
}

int SBEMU_GetChannels()
{
    return (SBEMU_MixerRegs[SBEMU_MIXERREG_MODEFILTER]&0x2) ? 2 : 1;
}

int SBEMU_GetSampleRate()
{
    return SBEMU_SampleRate;
}

int SBEMU_GetSampleBytes()
{
    return SBEMU_Samples + 1;
}

int SBEMU_GetAuto()
{
    return SBEMU_Auto;
}

int SBEMU_GetPos()
{
    return SBEMU_Pos;
}

int SBEMU_SetPos(int pos)
{
    return SBEMU_Pos = pos;
}

